;;;; AJL - geneticky algoritmus I, v4
;;; (C) MM, 2009

;;;; Zadanie: Numericka optimalizacia
;;;
;;;  Hlada sa extrem zadanej numerickej funkcie na danom intervale

;;;; Verzia 1 ******************************************************


;;; definicia globalnych premennych

(defparameter *body-length* 10)            ; dlzka binarneho retazca
(defparameter *population-size* 30)        ; velkost populacie
(defparameter *prob-crossover* 0.6)        ; pravdepodonost krizenia
(defparameter *prob-mutation* 0.05)        ; pravdepodobnost zmutovania (0<->1)
					   ;jedneho bitu binarneho retazca 
(defparameter *max-gen* 30)                ; pocet generacii evol. procesu
(defparameter *trace* nil)                 ; t - vypis stavu kazdu generaciu, 
                                           ; nil - vypis iba na konci

;;;  hlavna slucka evolucneho procesu 

(defun ga-run (funct)                      ; funct - funkcia pre optimalizaciu
  (let ((population (distribution (generate) funct))) ; inicializacia populacie
    (dotimes (i *max-gen*)                          ;  *max-gen* krat opakuj
      (print-population i population)               ;  vypis aktualneho stavu 
      (setf population (distribution (reproduce population) funct))
					            ; nova generacia
      )
    (print-population *max-gen* population)         ;  vypis koncoveho stavu 
    )
  'done
)

;;;  vypis aktualneho stavu populacie jedincov v tvare :
;;;  1. binarny retazec
;;;  2. cislo, ktore reprezentuje
;;;  3. vhodnost jedinca

(defun print-population (generation population)
  (when (or 
	 (= generation *max-gen*)
	 *trace*   
	 )
    (format t "~%~% ~D generation" generation)
    (mapc #'(lambda (x)
	      (format t "~%Jedinec ~D = ~D s vhodnostou ~D"
		      (cdr x) (string>num (cdr x)) (car x))
	      )
	  population)
    )
  )

;;;  inicializacia populacie  
;;;  vytvorenie *population-size* binarnych jedincov skladajucich sa 
;;;  zo samych 0, nasledne zmutovanych  s pravdepodobnostou 50 % 
;;;  pre kazdu poziciu binarneho retazca

(defun generate ()
  (let ((population))
    (dotimes (i *population-size*)
      (push (mutate 0.5 (make-list *body-length* :initial-element 0))
	    population
	    )
      )
    population
    )
  )

;;; nahrada povodnej populacie populaciou potomkov

(defun reproduce (population)
  (let ((offspring))                        ; definicia populacie potomkov 
                                            ; (spociatku prazdny zoznam)
    (dotimes (i (/ (length population) 2))  ; velkost-populacie/2 reprodukcii
      (let ((x (selectone population))      ; vyber rodica c.1.
	    (y (selectone population)))     ; vyber rodica c.2.
	(setf x (mutate *prob-mutation* x)) ; mutuj jedneho ...
	(setf y (mutate *prob-mutation* y)) ; ... a druheho ...
	(crossover x y)                     ; ... a kriz ich
	(setq offspring (nconc (list x y) offspring)) ;  pridaj potomkov (x y) 
					              ; do populacie potomkov
       )
      )
    offspring                               ; vysledna populacia potomkov   
    )
  )

;;; funkcia priradi jedincom vhodnost
;;; jedinec bude tvoreny 1 + *body-lenght* zoznamom:
;;;   - prvy prvok fitness
;;;   - *body-length* bitov binarneho retazca tela jedinca

(defun distribution (individuals funct)
  (mapcar #'(lambda (x) (cons (fitness x funct) x)) individuals)  
  )

;;; selekcia - vyber jedneho jedinca populacie binarnym turnajom

(defun selectone (distribution)           ;  populacia jedincov obsahujucich
					  ;  na prvej pozicii rel-fitness 
  (let* ((max (length distribution))           ; max - pocet jedincov 
	 (x (nth (random max) distribution))   ; nahodne vyber 1. jedinca
	 (y (nth (random max) distribution)))  ; nahodne vyber 2. jedinca
    (if (> (car x) (car y))                    ; porovnaj rel-fitness jedincov
	(cdr x)                                ; vrat vitaza ... 
	(cdr y)                                ; ... binarneho turnaja 
      )
    )
  )

;;; funkcia vykona crossover (1-bodove krizenie) chromozomov 
;;; jedincov x,y
;;; destruktivne meni vstupy !

(defun crossover (x y)
  (if (> (random 1.0) *prob-crossover*) ; crossover sa vykona 
                                        ; s pravdepodobnostou *prob-crossover*
    (list x y)                          ;  nil vetva, nevykona ziadnu zmenu
    (let* ((site (random (length x)))   ; nahodny vyber bodu prekrizenia
	   (swap (rest (nthcdr site x))))                               
      (setf (rest (nthcdr site x)) (rest (nthcdr site y)))        
      (setf (rest (nthcdr site x)) swap) 
      )
    )
  )

;;; funkcia vykona mutaciu chromozomu jedinca

(defun mutate (prob genotype)
  (mapcar #'(lambda (x)                   ; mapcar aplikuje funkciu lambda 
                                          ; na vs. prvky zoznamu
	      (if (> (random 1.0) prob) 
		x
		(if (= x 1) 
		  0
		  1)
		)
	      )
	  genotype
	  )
  )

;; funkcia vyrata vhodnost jedinca dosadenim odkodovaneho binarneho retazca 
;; do konkretnej funkcie vhodnosti 

(defun fitness (x funct)
  (let ((xarg (string>num x)))       ; xarg - cislo zakodovane bin. retazcom
    (funcall funct xarg)
    ) 
  )

;;; funkcia vykona priame premapovanie bin. retazca na kladne cislo 
;;; z intervalu <0,1>

(defun string>num (s)
  (let ((res 0)
	(l (length s)))
    (dotimes (i (length s))
      (setf res (+ res (* (nth (- l i 1) s) (expt 2 i))))   ; integer
      )
    (/ res (1- (expt 2 l)))                 ; normovane na interval <0,1>
    )
  )		

;;; spustenie celeho evolucneho procesu
;;;
;;;  (ga-run #'(lambda (x) (* (expt 2.7 (- x)) (sin (* 4 3.14 x)))))
;;;
;;;  (ga-run #'(lambda (x) (* x x))) 


